package suicideChess;

import java.util.ArrayList;
import java.util.Random;

import suicideChess.Board.NoPieceOnSquare;
import suicideChess.Square.NotAValidSquare;

/**
 * This class will contain all the AI.
 * @author Jean-Baptiste H&eacute;tier
 * @version $LastChangedRevision$, $LastChangedDate$
 */

public class ComputerPlayer {

    //best move found
    private static Move bestMove = null;
    
    //to tell the number of nodes searched
    private static int nodesSearched;
    
 
    /**
     * This asks the computer to compute a move
     * @param bitboard The current status of the {@link Board}
     * @return move A {@link Move}
     * @throws NotAValidSquare 
     * @see Board
     * @see Move
     */
    public static Move doRandomMove(Board bitboard) throws NotAValidSquare {
        Random generator = new Random();
        Rules.legalMovesForPlayer(bitboard);
        ArrayList<Move> allLegalMoves = Rules.getLegalMovesCapture();
        if (allLegalMoves.size()==0) {
            allLegalMoves = Rules.getLegalMovesNonCapture();
        }
        if (allLegalMoves.size()!=0) {
            return allLegalMoves.get(generator.nextInt(allLegalMoves.size()));
        }
        
        throw new RuntimeException("**Error** in doRandomMove");

    }
    
    /**
     * Basic MinMax
     * @param bitboard The bitboard
     * @return The best Move found 
     * @throws NotAValidSquare 
     * @throws NoPieceOnSquare 
     * @see Board
     * @see Move
     */
    public static Move doMinMaxMove(Board bitboard) throws NotAValidSquare, NoPieceOnSquare {
        bestMove = null;
        nodesSearched = 0;

        int bestScore = MinMax(bitboard, 0);
        if (SuicideChess.postThinkingOutput()) {
            System.out.println(SuicideChess.PLY_DEPTH+" "+bestScore*100+" 0 "+nodesSearched+" "+bestMove);
        }
        return bestMove;
    }
    
    
    private static int MinMax(Board bitboard, int currentDepth) throws NotAValidSquare, NoPieceOnSquare {
        if (currentDepth >= SuicideChess.PLY_DEPTH) {
            return bitboard.getBoardValue();
        }
        
        nodesSearched++;
        
        Rules.legalMovesForPlayer(bitboard);
        ArrayList<Move> allLegalMoves = Rules.getLegalMovesCapture();
        if (allLegalMoves.size()==0) {
            allLegalMoves = Rules.getLegalMovesNonCapture();
        }
        if (allLegalMoves.size()==0) {
            if (bitboard.getCurrentPlayer()==Piece.BLACK) {
                return Board.BLACK_WINS;
            } else {
                return Board.WHITE_WINS;
            }
        } else {
            ArrayList<Integer> bestMoveIndex=new ArrayList<Integer>();
            int currentScore;
            int bestScoreSoFar;
            if (bitboard.getCurrentPlayer()==Piece.BLACK) {            
                bestScoreSoFar=Board.WHITE_WINS+1; //any move even a WHITE_WINS will be better than that
                for (int i=0; i<allLegalMoves.size(); i++) {
                    Board boardCopy = new Board(bitboard);
                    boardCopy.doMove(allLegalMoves.get(i));
                    currentScore=MinMax(boardCopy,currentDepth+1);
                    if (currentScore <= bestScoreSoFar) { //black tries to minimise his score
                        if (currentScore<bestScoreSoFar) {
                            bestScoreSoFar = currentScore;
                            bestMoveIndex.clear();
                        }
                        bestMoveIndex.add(i);
                    }
                }
            } else { //white piece
                bestScoreSoFar=Board.BLACK_WINS-1; //any move even a BLACK_WINS will be better than that
                for (int i=0; i<allLegalMoves.size(); i++) {
                    Board boardCopy = new Board(bitboard);
                    boardCopy.doMove(allLegalMoves.get(i));
                    currentScore=MinMax(boardCopy,currentDepth+1);
                    if (currentScore >= bestScoreSoFar) { //white tries to maximise his score
                        if (currentScore>bestScoreSoFar) {
                            bestScoreSoFar = currentScore;
                            bestMoveIndex.clear();
                        }
                        bestMoveIndex.add(i);
                    }
                }
            }
            
            //select one of the best moves randomly
            Random generator = new Random();
            if (currentDepth == 0) { System.out.println(bestMoveIndex.size()); }
            bestMove = allLegalMoves.get(bestMoveIndex.get(generator.nextInt(bestMoveIndex.size())));
            
            return bestScoreSoFar;
        }  
    }
        
    /**
     * Alpha-Beta
     * @param bitboard The bitboard
     * @return The best Move found 
     * @throws NotAValidSquare 
     * @throws NoPieceOnSquare 
     * @see Board
     * @see Move
     */
    public static Move doAlphaBetaMove(Board bitboard) throws NotAValidSquare, NoPieceOnSquare {
        bestMove = null;
        nodesSearched = 0;
 
        int bestScore = AlphaBeta(bitboard, 0, Board.BLACK_WINS-1, Board.WHITE_WINS+1);
        if (SuicideChess.postThinkingOutput()) {
            System.out.println(SuicideChess.PLY_DEPTH+" "+bestScore*100+" 0 "+nodesSearched+" "+bestMove);
        } 
        
        return bestMove;
    }
    
    
    private static int AlphaBeta(Board bitboard, int currentDepth, int alpha, int beta) throws NotAValidSquare, NoPieceOnSquare {
        if (currentDepth >= SuicideChess.PLY_DEPTH) {
            return bitboard.getBoardValue();
        }
        
        nodesSearched++;
        //System.out.println(currentDepth + " : " + alpha + " " + beta);       
        
        Rules.legalMovesForPlayer(bitboard);
        ArrayList<Move> allLegalMoves = Rules.getLegalMovesCapture();
        if (allLegalMoves.size()==0) {
            allLegalMoves = Rules.getLegalMovesNonCapture();
        }
        if (allLegalMoves.size()==0) {
            if (bitboard.getCurrentPlayer()==Piece.BLACK) {
                return Board.BLACK_WINS;
            } else {
                return Board.WHITE_WINS;
            }
        
        } else {
            ArrayList<Integer> bestMoveIndex=new ArrayList<Integer>();
            int bestScoreSoFar;
            int currentScore;
            if (bitboard.getCurrentPlayer()==Piece.BLACK) {            
                bestScoreSoFar=Board.WHITE_WINS+1; //any move even a WHITE_WINS will be better than that
                for (int i=0; i<allLegalMoves.size(); i++) {
                    Board boardCopy = new Board(bitboard);
                    boardCopy.doMove(allLegalMoves.get(i));
                    currentScore=AlphaBeta(boardCopy,currentDepth+1,alpha,beta);
                    beta = Math.min(beta,currentScore);
                    if ((currentScore <= bestScoreSoFar)) { //black tries to minimise his score
                        if (currentScore<bestScoreSoFar) {
                            bestScoreSoFar = currentScore;
                            bestMoveIndex.clear();
                        }
                        bestMoveIndex.add(i);
                    }

                    
                    //select one of the best moves randomly
                    Random generator = new Random();
                    //if (currentDepth == 0) { System.out.println(bestMoveIndex.size()); }
                    bestMove = allLegalMoves.get(bestMoveIndex.get(generator.nextInt(bestMoveIndex.size())));

                    //System.out.println(currentDepth + " : " + alpha + " " + beta);       
                    
                    if(beta<=alpha) {
                        return alpha; //pruning
                    }
                }
                return beta;
            } else { //white piece
                bestScoreSoFar=Board.BLACK_WINS-1; //any move even a BLACK_WINS will be better than that
                //alpha, highest score seen so far by Max in minmax
                alpha = Board.BLACK_WINS-1;

                for (int i=0; i<allLegalMoves.size(); i++) {
                    Board boardCopy = new Board(bitboard);
                    boardCopy.doMove(allLegalMoves.get(i));
                    currentScore=AlphaBeta(boardCopy,currentDepth+1,alpha,beta);
                    alpha = Math.max(alpha,currentScore);
                    if ((currentScore >= bestScoreSoFar)) { //white tries to maximise his score
                        if (currentScore>bestScoreSoFar) {
                            bestScoreSoFar = currentScore;
                            bestMoveIndex.clear();
                        }
                        bestMoveIndex.add(i);
                    }

                    
                    //select one of the best moves randomly
                    Random generator = new Random();
                    //if (currentDepth == 0) { System.out.println(bestMoveIndex.size()); }
                    bestMove = allLegalMoves.get(bestMoveIndex.get(generator.nextInt(bestMoveIndex.size())));
                    
                    //System.out.println(currentDepth + " : " + alpha + " " + beta);       
                    
                    if(alpha>=beta) {
                        return beta; //pruning
                    }   
                }
                return alpha;
            }
        }  
    }
 
}